/*
 * RSH - Robte si hocičo
 *
 * Ktokoľvek a čokoľvek smie kopírovať, upravovať, zverejňovať,
 * používať, kompilovať, predávať a šíriť toto dielo
 * a všetky jeho časti v akejkoľvek podobe na akékoľvek účely,
 * komerčné a nekomerčné, bez akýchkoľvek obmedzení,
 * bez dodržiavania akýchkoľvek podmienok
 * a akýmikoľvek prostriedkami.
 *
 * Jeder und alles darf kopieren, bearbeiten, veröffentlichen,
 * verwenden, kompilieren, verkaufen und verbreiten dieses Werk
 * und alle seine Teile in irgendeiner Form für irgendeinen Zweck,
 * kommerziell und nichtkommerziell, ohne Beschränkungen,
 * ohne irgendwelche Bedingungen einzuhalten
 * und mit allen Mitteln.
 *
 * Anyone and anything may copy, edit, publish,
 * use, compile, sell and distribute this work
 * and all its parts in any form for any purpose,
 * commercial and non-commercial, without any restrictions,
 * without complying with any conditions
 * and by any means.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "genk.h"
#include "parser.h"

static void genkZlyhanie(char *text)
{
    printf("(GENK) Zlyhanie!!!\n");
    printf(text);
    exit(0);
}

/* Externee premennee. */
typedef struct _extprem {
    char *meno;
    struct _extprem *extprem;
} EXTPREM;

static EXTPREM *extpremDeklaruj(EXTPREM *staraa, char *meno)
{
    EXTPREM *extprem;

    for (extprem = staraa; extprem; extprem = extprem->extprem)
    {
        if (strcmp(meno, extprem->meno) == 0)
        {
            genkZlyhanie("DOROBITH extpremDeklaruj+++\n");
        }
    }

    extprem = malloc(sizeof(EXTPREM));

    if (extprem == NULL) genkZlyhanie("Nepodarilo sa alokovath pameth\n");
    extprem->meno = malloc(strlen(meno) + 1);

    if (extprem->meno == NULL) genkZlyhanie("Nepodarilo sa alokovath pameth\n");
    strcpy(extprem->meno, meno);

    extprem->extprem = staraa;

    return (extprem);
}

static char *extpremReferencia(EXTPREM *extprem, char *meno)
{
    for (; extprem; extprem = extprem->extprem)
    {
        if (strcmp(meno, extprem->meno) == 0)
        {
            return (extprem->meno);
        }
    }

    printf("meno: %s\n", meno);
    genkZlyhanie("Referencia neexistujuucej ext. premennej\n");

    return (NULL);
}

/* Lokaalne premennee. */
typedef struct _prem {
    char *meno;
    int index; /* Vzdialenosth od EBP. */
    struct _prem *prem;
} PREM;

static int prem_index = 0;
static int param_index = 4;

static PREM *vytvorPrem(PREM *staraa, char *meno, PREM *prem_limit)
{
    PREM *prem;

    /* prem_limit sluuzhi na oddelenie premennyych {} vyyrokov. */
    for (prem = staraa; prem != prem_limit; prem = prem->prem)
    {
        if (strcmp(meno, prem->meno) == 0)
        {
            genkZlyhanie("Redeklaraacia premennej\n");
        }
    }

    prem = malloc(sizeof(PREM));

    if (prem == NULL) genkZlyhanie("Nepodarilo sa alokovath pameth\n");
    prem->meno = malloc(strlen(meno) + 1);

    if (prem->meno == NULL) genkZlyhanie("Nepodarilo sa alokovath pameth\n");
    strcpy(prem->meno, meno);

    prem->index = prem_index -= 4;
    prem->prem = staraa;

    return (prem);
}

static PREM *vytvorParam(PREM *staraa, char *meno, PREM *prem_limit)
{
    PREM *prem;

    /* prem_limit sluuzhi na oddelenie premennyych {} vyyrokov. */
    for (prem = staraa; prem != prem_limit; prem = prem->prem)
    {
        if (strcmp(meno, prem->meno) == 0)
        {
            genkZlyhanie("Redeklaraacia premennej\n");
        }
    }

    prem = malloc(sizeof(PREM));

    if (prem == NULL) genkZlyhanie("Nepodarilo sa alokovath pameth\n");
    prem->meno = malloc(strlen(meno) + 1);

    if (prem->meno == NULL) genkZlyhanie("Nepodarilo sa alokovath pameth\n");
    strcpy(prem->meno, meno);

    prem->index = param_index += 4;
    if (staraa)
    {
        prem->prem = staraa;
    }
    else
    {
        prem->prem = NULL;
    }

    return (prem);
}

static int indexPrem(PREM *prem, char *meno)
{
    for (; prem; prem = prem->prem)
    {
        if (strcmp(meno, prem->meno) == 0)
        {
            return (prem->index);
        }
    }

    return (0);
}

static int znicshPrem(PREM *prem, PREM *prem_limit)
{
    PREM *staraa;
    int pocshet = 0;

    /* prem_limit sluuzhi na oddelenie premennyych {} vyyrokov. */
    for (; prem != prem_limit; )
    {
        staraa = prem;
        prem = prem->prem;

        if (staraa->index < 0) pocshet += 4;
        free(staraa);
    }

    prem_index += pocshet;

    return (pocshet);
}

static EXTPREM *extprem = NULL;
static PREM *prem = NULL;
static PREM *prem_limit = NULL;
static int chislo_pre_znachky = 1;
static int aktuaalny_cyklus = 0;

static void genkAst(AST *ast, FILE *f)
{
    if (ast == NULL) return;

    switch (ast->typ) {
        case TranslationUnit: {
            genkAst(ast->un.TranslationUnit.translationUnit, f);
            genkAst(ast->un.TranslationUnit.funkcia, f);

            break;
        }

        case Funkcia: {
            AST_LIST *ast_list;

            fprintf(f, "\t.globl _%s\n", ast->un.Funkcia.meno);
            fprintf(f, "_%s:\n", ast->un.Funkcia.meno);
            /* Proloog funkcie. */
            fprintf(f, "\tpushl\t%%ebp\n");
            fprintf(f, "\tmovl\t%%esp, %%ebp\n");

            for (ast_list = ast->un.Funkcia.parameter_list;
                 ast_list;
                 ast_list = ast_list->ast_list)
            {
                genkAst(ast_list->ast, f);
            }

            for (ast_list = ast->un.Funkcia.blok_item_list;
                 ast_list;
                 ast_list = ast_list->ast_list)
            {
                genkAst(ast_list->ast, f);
            }

            /* main() maa vraatith 0 aj bez returnu +++. */
            fprintf(f, "\tmovl\t%%ebp, %%esp\n");
            fprintf(f, "\tpopl\t%%ebp\n");
            fprintf(f, "\tmovl\t$0, %%eax\n");
            fprintf(f, "\tret\n");

            break;
        }

        case Parameter: {
            prem = vytvorParam(prem, ast->un.Parameter.meno, prem_limit);

            break;
        }

        case ExternalDekl: {
            extprem = extpremDeklaruj(extprem, ast->un.ExternalDekl.meno);

            break;
        }

        case If: {
            int aktuaalne_chislo = chislo_pre_znachky;
            chislo_pre_znachky++;

            genkAst(ast->un.If.exp, f);

            fprintf(f, "\tcmpl\t$0, %%eax\n");
            if (ast->un.If.else_statement)
            {
                fprintf(f, "\tjne\t_moznosth_1_%i\n", aktuaalne_chislo);
                fprintf(f, "\tjmp\t_moznosth_2_%i\n", aktuaalne_chislo);
                fprintf(f, "_moznosth_1_%i:\n", aktuaalne_chislo);
            }
            else
            {
                fprintf(f, "\tjne\t_moznosth_%i\n", aktuaalne_chislo);
                fprintf(f, "\tjmp\t_koniec_%i\n", aktuaalne_chislo);
                fprintf(f, "_moznosth_%i:\n", aktuaalne_chislo);
            }

            genkAst(ast->un.If.statement, f);

            if (ast->un.If.else_statement)
            {
                fprintf(f, "\tjmp\t_koniec_%i\n", aktuaalne_chislo);

                fprintf(f, "_moznosth_2_%i:\n", aktuaalne_chislo);
                genkAst(ast->un.If.else_statement, f);
            }

            fprintf(f, "_koniec_%i:\n", aktuaalne_chislo);

            break;
        }

        case Return: {
            genkAst(ast->un.Return.exp, f);

            /* Epiloog funkcie. */
            fprintf(f, "\tmovl\t%%ebp, %%esp\n");
            fprintf(f, "\tpopl\t%%ebp\n");
            fprintf(f, "\tret\n");

            break;
        }

        case Blok: {
            AST_LIST *ast_list;
            PREM *ulozhenyy_prem_limit = prem_limit;
            int esp_zmena;

            prem_limit = prem;

            for (ast_list = ast->un.Blok.blok_item_list;
                 ast_list;
                 ast_list = ast_list->ast_list)
            {
                genkAst(ast_list->ast, f);
            }

            esp_zmena = znicshPrem(prem, prem_limit);
            if (esp_zmena) fprintf(f, "\taddl\t$%i, %%esp\n", esp_zmena);

            prem_limit = ulozhenyy_prem_limit;

            break;
        }

        case For: {
            PREM *ulozhenyy_prem_limit = prem_limit;
            int esp_zmena;
            int aktuaalne_chislo = chislo_pre_znachky;
            int ulozhenyy_aktuaalny_cyklus = aktuaalny_cyklus;

            aktuaalny_cyklus = aktuaalne_chislo;
            chislo_pre_znachky++;

            prem_limit = prem;

            if (ast->un.For.dob_exp1) genkAst(ast->un.For.dob_exp1, f);

            fprintf(f, "_zhacshiatok_%i:\n", aktuaalne_chislo);

            genkAst(ast->un.For.exp2, f);
            fprintf(f, "\tcmpl\t$0, %%eax\n");
            fprintf(f, "\tjne\t_telo_%i\n", aktuaalne_chislo);
            fprintf(f, "\tjmp\t_koniec_%i\n", aktuaalne_chislo);
            fprintf(f, "_telo_%i:\n", aktuaalne_chislo);

            genkAst(ast->un.For.statement, f);

            fprintf(f, "_continue_%i:\n", aktuaalne_chislo);

            if (ast->un.For.dob_exp3) genkAst(ast->un.For.dob_exp3, f);
            fprintf(f, "\tjmp\t_zhacshiatok_%i\n", aktuaalne_chislo);
            fprintf(f, "_koniec_%i:\n", aktuaalne_chislo);

            aktuaalny_cyklus = ulozhenyy_aktuaalny_cyklus;

            esp_zmena = znicshPrem(prem, prem_limit);
            if (esp_zmena) fprintf(f, "\taddl\t$%i, %%esp\n", esp_zmena);

            prem_limit = ulozhenyy_prem_limit;

            break;
        }

        case While: {
            int ulozhenyy_aktuaalny_cyklus = aktuaalny_cyklus;
            int aktuaalne_chislo = chislo_pre_znachky;

            aktuaalny_cyklus = aktuaalne_chislo;
            chislo_pre_znachky++;

            fprintf(f, "_zhacshiatok_%i:\n", aktuaalne_chislo);

            genkAst(ast->un.While.exp, f);
            fprintf(f, "\tcmpl\t$0, %%eax\n");
            fprintf(f, "\tjne\t_telo_%i\n", aktuaalne_chislo);
            fprintf(f, "\tjmp\t_koniec_%i\n", aktuaalne_chislo);
            fprintf(f, "_telo_%i:\n", aktuaalne_chislo);

            genkAst(ast->un.While.statement, f);

            fprintf(f, "_continue_%i:\n", aktuaalne_chislo);
            fprintf(f, "\tjmp\t_zhacshiatok_%i\n", aktuaalne_chislo);
            fprintf(f, "_koniec_%i:\n", aktuaalne_chislo);

            aktuaalny_cyklus = ulozhenyy_aktuaalny_cyklus;

            break;
        }

        case Do_While: {
            int ulozhenyy_aktuaalny_cyklus = aktuaalny_cyklus;
            int aktuaalne_chislo = chislo_pre_znachky;

            aktuaalny_cyklus = aktuaalne_chislo;
            chislo_pre_znachky++;

            fprintf(f, "_zhacshiatok_%i:\n", aktuaalne_chislo);

            genkAst(ast->un.Do_While.statement, f);

            genkAst(ast->un.Do_While.exp, f);
            fprintf(f, "\tcmpl\t$0, %%eax\n");
            fprintf(f, "\tjne\t_continue_%i\n", aktuaalne_chislo);
            fprintf(f, "\tjmp\t_koniec_%i\n", aktuaalne_chislo);

            fprintf(f, "_continue_%i:\n", aktuaalne_chislo);
            fprintf(f, "\tjmp\t_zhacshiatok_%i\n", aktuaalne_chislo);
            fprintf(f, "_koniec_%i:\n", aktuaalne_chislo);

            aktuaalny_cyklus = ulozhenyy_aktuaalny_cyklus;

            break;
        }

        case Continue: {
            if (aktuaalny_cyklus == 0) genkZlyhanie("CONTINUE mimo cyklu\n");
            fprintf(f, "\tjmp\t_chvost_%i\n", aktuaalny_cyklus);

            break;
        }

        case Break: {
            if (aktuaalny_cyklus == 0) genkZlyhanie("BREAK mimo cyklu\n");
            fprintf(f, "\tjmp\t_koniec_%i\n", aktuaalny_cyklus);

            break;
        }

        case Dekl: {
            prem = vytvorPrem(prem, ast->un.Dekl.meno, prem_limit);

            if (ast->un.Dekl.dob_exp) genkAst(ast->un.Dekl.dob_exp, f);
            fprintf(f, "\tpushl\t%%eax\n");

            break;
        }

        case UnOp: {
            switch (ast->un.UnOp.typ) {
                case UNAARNA_OP_AMPERSAND: {
                    if (ast->un.UnOp.exp->typ == Prem)
                    {
                        int index = indexPrem(prem,
                                              ast->un.UnOp.exp->un.Prem);

                        if (index)
                        {
                            fprintf(f, "\tmovl\t%%ebp, %%eax\n");
                            fprintf(f, "\taddl\t$%i, %%eax\n", index);
                        }
                        else
                        {
                            fprintf(f, "\tmovl\t$_%s, %%eax\n",
                                    extpremReferencia(extprem,
                                                      ast->un.UnOp.exp
                                                      ->un.Prem));
                        }
                    }
                    else genkZlyhanie("Lhavaa strana neimplementovanaa!\n");

                    break;
                }

                case UNAARNA_OP_NEG: {
                    genkAst(ast->un.UnOp.exp, f);
                    fprintf(f, "\tneg\t%%eax\n");

                    break;
                }

                case UNAARNA_OP_BITCOMPLEMENT: {
                    genkAst(ast->un.UnOp.exp, f);
                    fprintf(f, "\tnot\t%%eax\n");

                    break;
                }

                case UNAARNA_OP_LOGNEG: {
                    genkAst(ast->un.UnOp.exp, f);
                    fprintf(f, "\tcmpl\t$0, %%eax\n");
                    fprintf(f, "\tmovl\t$0, %%eax\n");
                    fprintf(f, "\tsete\t%%al\n");

                    break;
                }
            }

            break;
        }

        case BinOp: {
            genkAst(ast->un.BinOp.exp1, f);

            switch (ast->un.BinOp.typ) {
                case BINAARNA_OP_HVIEZDA:
                    fprintf(f, "\tpushl\t%%eax\n");
                    genkAst(ast->un.BinOp.exp2, f);

                    fprintf(f, "\tpopl\t%%ecx\n");
                    fprintf(f, "\timull\t%%ecx, %%eax\n");
                    break;

                case BINAARNA_OP_LOMKA:
                    fprintf(f, "\tpushl\t%%eax\n");
                    genkAst(ast->un.BinOp.exp2, f);

                    fprintf(f, "\tmovl\t%%eax, %%ecx\n");
                    fprintf(f, "\tpopl\t%%eax\n");
                    fprintf(f, "\txorl\t%%edx, %%edx\n");
                    fprintf(f, "\tidivl\t%%ecx, %%eax\n");
                    break;

                case BINAARNA_OP_PLUS:
                    fprintf(f, "\tpushl\t%%eax\n");
                    genkAst(ast->un.BinOp.exp2, f);

                    fprintf(f, "\tpopl\t%%ecx\n");
                    fprintf(f, "\taddl\t%%ecx, %%eax\n");
                    break;

                case BINAARNA_OP_MIINUS:
                    fprintf(f, "\tpushl\t%%eax\n");
                    genkAst(ast->un.BinOp.exp2, f);

                    fprintf(f, "\tmovl\t%%eax, %%ecx\n");
                    fprintf(f, "\tpopl\t%%eax\n");
                    fprintf(f, "\tsubl\t%%ecx, %%eax\n");
                    break;

                case BINAARNA_OP_LESS:
                    fprintf(f, "\tpushl\t%%eax\n");
                    genkAst(ast->un.BinOp.exp2, f);

                    fprintf(f, "\tpopl\t%%ecx\n");
                    fprintf(f, "\tcmpl\t%%eax, %%ecx\n");
                    fprintf(f, "\tmovl\t$0, %%eax\n");
                    fprintf(f, "\tsetl\t%%al\n");
                    break;

                case BINAARNA_OP_LESS_EQUAL:
                    fprintf(f, "\tpushl\t%%eax\n");
                    genkAst(ast->un.BinOp.exp2, f);

                    fprintf(f, "\tpopl\t%%ecx\n");
                    fprintf(f, "\tcmpl\t%%eax, %%ecx\n");
                    fprintf(f, "\tmovl\t$0, %%eax\n");
                    fprintf(f, "\tsetle\t%%al\n");
                    break;

                case BINAARNA_OP_GREATER:
                    fprintf(f, "\tpushl\t%%eax\n");
                    genkAst(ast->un.BinOp.exp2, f);

                    fprintf(f, "\tpopl\t%%ecx\n");
                    fprintf(f, "\tcmpl\t%%eax, %%ecx\n");
                    fprintf(f, "\tmovl\t$0, %%eax\n");
                    fprintf(f, "\tsetg\t%%al\n");
                    break;

                case BINAARNA_OP_GREATER_EQUAL:
                    fprintf(f, "\tpushl\t%%eax\n");
                    genkAst(ast->un.BinOp.exp2, f);

                    fprintf(f, "\tpopl\t%%ecx\n");
                    fprintf(f, "\tcmpl\t%%eax, %%ecx\n");
                    fprintf(f, "\tmovl\t$0, %%eax\n");
                    fprintf(f, "\tsetge\t%%al\n");
                    break;

                case BINAARNA_OP_EQUAL:
                    fprintf(f, "\tpushl\t%%eax\n");
                    genkAst(ast->un.BinOp.exp2, f);

                    fprintf(f, "\tpopl\t%%ecx\n");
                    fprintf(f, "\tcmpl\t%%eax, %%ecx\n");
                    fprintf(f, "\tmovl\t$0, %%eax\n");
                    fprintf(f, "\tsete\t%%al\n");
                    break;

                case BINAARNA_OP_NOT_EQUAL:
                    fprintf(f, "\tpushl\t%%eax\n");
                    genkAst(ast->un.BinOp.exp2, f);

                    fprintf(f, "\tpopl\t%%ecx\n");
                    fprintf(f, "\tcmpl\t%%eax, %%ecx\n");
                    fprintf(f, "\tmovl\t$0, %%eax\n");
                    fprintf(f, "\tsetne\t%%al\n");
                    break;

                case BINAARNA_OP_LOG_AND: {
                    int aktuaalne_chislo = chislo_pre_znachky;
                    chislo_pre_znachky++;

                    fprintf(f, "\tcmpl\t$0, %%eax\n");
                    fprintf(f, "\tjne\t_moznosth%i\n", aktuaalne_chislo);
                    fprintf(f, "\tjmp\t_koniec%i\n", aktuaalne_chislo);
                    fprintf(f, "_moznosth%i:\n", aktuaalne_chislo);

                    genkAst(ast->un.BinOp.exp2, f);
                    fprintf(f, "\tcmpl\t$0, %%eax\n");
                    fprintf(f, "\tmovl\t$0, %%eax\n");
                    fprintf(f, "\tsetne\t%%al\n");
                    fprintf(f, "_koniec%i:\n", aktuaalne_chislo);

                    break;
                }

                case BINAARNA_OP_LOG_OR: {
                    int aktuaalne_chislo = chislo_pre_znachky;
                    chislo_pre_znachky++;

                    fprintf(f, "\tcmpl\t$0, %%eax\n");
                    fprintf(f, "\tje\t_moznosth%i\n", aktuaalne_chislo);
                    fprintf(f, "\tmovl\t$1, %%eax\n");
                    fprintf(f, "\tjmp\t_koniec%i\n", aktuaalne_chislo);
                    fprintf(f, "_moznosth%i:\n", aktuaalne_chislo);

                    genkAst(ast->un.BinOp.exp2, f);
                    fprintf(f, "\tcmpl\t$0, %%eax\n");
                    fprintf(f, "\tmovl\t$0, %%eax\n");
                    fprintf(f, "\tsetne\t%%al\n");
                    fprintf(f, "_koniec%i:\n", aktuaalne_chislo);

                    break;
                }

                default: genkZlyhanie("Chyybajuuca BinOp\n");
            }

            break;
        }

        case TernOp: {
            int aktuaalne_chislo = chislo_pre_znachky;
            chislo_pre_znachky++;

            genkAst(ast->un.TernOp.exp1, f);
            fprintf(f, "\tcmpl\t$0, %%eax\n");
            fprintf(f, "\tjne\t_moznosth_1_%i\n", aktuaalne_chislo);
            fprintf(f, "\tjmp\t_moznosth_2_%i\n", aktuaalne_chislo);

            fprintf(f, "_moznosth_1_%i:\n", aktuaalne_chislo);
            genkAst(ast->un.TernOp.exp2, f);
            fprintf(f, "\tjmp\t_koniec_%i\n", aktuaalne_chislo);

            fprintf(f, "_moznosth_2_%i:\n", aktuaalne_chislo);
            genkAst(ast->un.TernOp.exp3, f);

            fprintf(f, "_koniec_%i:\n", aktuaalne_chislo);

            break;
        }

        case Assign: {
            if (ast->un.Assign.exp1->typ == Prem)
            {
                int index = indexPrem(prem, ast->un.Assign.exp1->un.Prem);

                genkAst(ast->un.Assign.exp2, f);

                if (index)
                {
                    fprintf(f, "\tmovl\t%%eax, %i(%%ebp)\n", index);
                }
                else
                {
                    fprintf(f, "\tmovl\t%%eax, _%s\n",
                            extpremReferencia(extprem,
                                              ast->un.Assign.exp1->un.Prem));
                }
            }
            else genkZlyhanie("Lhavaa strana neimplementovanaa!\n");

            break;
        }

        case Volanie: {
            AST_LIST *ast_list;
            AST_LIST *dhalej;
            AST_LIST *minul = NULL;
            int index = 0;

            /* Parametre sa tlacshia zprava do lhava,
             * teda sa musii ich poradie prevraatith. */
            for (ast_list = ast->un.Volanie.assign_exp_list;
                 ast_list;
                 ast_list = dhalej)
            {
                dhalej = ast_list->ast_list;
                ast_list->ast_list = minul;
                minul = ast_list;
                ast->un.Volanie.assign_exp_list = ast_list;
            }

            for (ast_list = ast->un.Volanie.assign_exp_list;
                 ast_list;
                 ast_list = ast_list->ast_list)
            {
                genkAst(ast_list->ast, f);
                fprintf(f, "\tpushl\t%%eax\n");
                index += 4;
            }
            fprintf(f, "\tcall\t_%s\n", ast->un.Volanie.meno);
            if (index) fprintf(f, "\taddl\t$%i, %%esp\n", index);

            break;
        }

        case Prem: {
            int index = indexPrem(prem, ast->un.Prem);
            if (index)
            {
                fprintf(f, "\tmovl\t%i(%%ebp), %%eax\n", index);
            }
            else
            {
                fprintf(f, "\tmovl\t_%s, %%eax\n",
                        extpremReferencia(extprem, ast->un.Prem));
            }

            break;
        }

        case Konshtanta: {
            fprintf(f, "\tmovl\t$%i, %%eax\n", ast->un.Konshtanta);
            break;
        }
    }
}

void genkGeneruj(AST *ast, FILE *f)
{
    genkAst(ast, f);

    if (extprem)
    {
        EXTPREM *staraa;

        fprintf(f, "\n\t.data\n");
        for (; extprem; )
        {
            fprintf(f, "\t.globl _%s\n", extprem->meno);
            fprintf(f, "\t.align 2\n");
            fprintf(f, "_%s:\n", extprem->meno);
            fprintf(f, "\t.long 0\n");

            staraa = extprem;
            extprem = extprem->extprem;

            free(staraa);
        }
    }
}
