/*
 * RSH - Robte si hocičo
 *
 * Ktokoľvek a čokoľvek smie kopírovať, upravovať, zverejňovať,
 * používať, kompilovať, predávať a šíriť toto dielo
 * a všetky jeho časti v akejkoľvek podobe na akékoľvek účely,
 * komerčné a nekomerčné, bez akýchkoľvek obmedzení,
 * bez dodržiavania akýchkoľvek podmienok
 * a akýmikoľvek prostriedkami.
 *
 * Jeder und alles darf kopieren, bearbeiten, veröffentlichen,
 * verwenden, kompilieren, verkaufen und verbreiten dieses Werk
 * und alle seine Teile in irgendeiner Form für irgendeinen Zweck,
 * kommerziell und nichtkommerziell, ohne Beschränkungen,
 * ohne irgendwelche Bedingungen einzuhalten
 * und mit allen Mitteln.
 *
 * Anyone and anything may copy, edit, publish,
 * use, compile, sell and distribute this work
 * and all its parts in any form for any purpose,
 * commercial and non-commercial, without any restrictions,
 * without complying with any conditions
 * and by any means.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include "parser.h"
#include "lexer.h"

static void parserZlyhanie(char *text)
{
    printf("(PARSER) Zlyhanie!!!\n");
    printf(text);
    exit(0);
}

typedef struct _parser_tok {
    int tok;
    char *tokHodnota;
    struct _parser_tok *parser_tok;
} PARSER_TOK;

static char tokHodnota[MAX_IDENT];
static PARSER_TOK *parser_tok = NULL;

static int parserLex(FILE *f)
{
    int tok;

    if (parser_tok == NULL)
    {
        tok = lexerLex(f);
        strcpy(tokHodnota, lexerHodnota());
    }
    else
    {
        tok = parser_tok->tok;
        if (parser_tok->tokHodnota)
        {
            strcpy(tokHodnota, parser_tok->tokHodnota);
            free(parser_tok->tokHodnota);
        }
        {
            PARSER_TOK *novyy_parser_tok = parser_tok->parser_tok;

            free(parser_tok);
            parser_tok = novyy_parser_tok;
        }
    }

    return (tok);
}

static int parserPozriLex(FILE *f)
{
    int tok;

    if (parser_tok == NULL)
    {
        parser_tok = malloc(sizeof(PARSER_TOK));
        if (parser_tok == NULL) parserZlyhanie("PARSER_TOK alok.\n");
        parser_tok->parser_tok = NULL;

        tok = lexerLex(f);
        parser_tok->tok = tok;

        if (lexerHodnota()[0] != '\0')
        {
            parser_tok->tokHodnota = malloc(strlen(lexerHodnota()) + 1);
            if (parser_tok->tokHodnota == NULL)
            {
                parserZlyhanie("parser_tok->tokHodnota alok.\n");
            }
            strcpy(parser_tok->tokHodnota, lexerHodnota());
            strcpy(tokHodnota, parser_tok->tokHodnota);
        }
        else parser_tok->tokHodnota = NULL;
    }
    else
    {
        tok = parser_tok->tok;
        if (parser_tok->tokHodnota)
        {
            strcpy(tokHodnota, parser_tok->tokHodnota);
        }
    }

    return (tok);
}

static char *parserLexHodnota(void)
{
    return (tokHodnota);
}

static void necshakanyyTok(int tok)
{
    lexerTest(tok);
    printf("hodnota: %s\n", parserLexHodnota());
    parserZlyhanie("Necshakany tok.\n");
}

static void cshakajTok(int cshakanyy_tok, FILE *f)
{
    int tok = parserLex(f);

    if (tok != cshakanyy_tok)
    {
        printf("(PARSER) Zlyhanie!!!\n");
        printf("Ocshakaavanyy tok.: ");
        lexerTest(cshakanyy_tok);
        printf("Namiesto toho je tok.: ");
        lexerTest(tok);
        exit(0);
    }
}


static AST *alokujAst(int typ)
{
    AST *ast = malloc(sizeof(AST));

    if (ast == NULL) parserZlyhanie("Nepodarilo sa alokovath pameth\n");
    ast->typ = typ;

    return (ast);
}

static AST_LIST *vytvorAstList(AST *ast)
{
    AST_LIST *ast_list = malloc(sizeof(AST_LIST));

    if (ast_list == NULL) parserZlyhanie("Nepodarilo sa alokovath pameth\n");
    ast_list->ast = ast;
    ast_list->ast_list = NULL;

    return (ast_list);
}

static char *identMeno()
{
    char *meno = malloc(strlen(parserLexHodnota()) + 1);

    if (meno == NULL) parserZlyhanie("Nepodarilo sa alokovath pameth\n");
    strcpy(meno, parserLexHodnota());

    return (meno);
}

static AST *parserAssignExp(FILE *f);
static AST *parserExp(FILE *f);
static AST *parserDekl(FILE *f);
static AST *parserBlokItem(FILE *f);

static AST *parserPrimaryExp(FILE *f)
{
    AST *ast;
    int tok;

    tok = parserLex(f);

    switch (tok) {
        case INTLIT: {
            ast = alokujAst(Konshtanta);
            ast->un.Konshtanta = (int)strtol(parserLexHodnota(), NULL, 10);

            break;
        }

        case LPAREN: {
            ast = parserExp(f);
            cshakajTok(RPAREN, f);

            break;
        }

        case IDENT: {
            char *meno = identMeno();

            /* +++Tu by to chcelo rozoznath funkciu od premennej. */
            if (parserPozriLex(f) == LPAREN)
            {
                parserLex(f);

                ast = alokujAst(Volanie);
                ast->un.Volanie.meno = meno;

                if (parserPozriLex(f) == RPAREN)
                {
                    ast->un.Volanie.assign_exp_list = NULL;
                    parserLex(f);
                }
                else
                {
                    AST_LIST *ast_list;

                    ast_list = vytvorAstList(parserAssignExp(f));
                    ast->un.Volanie.assign_exp_list = ast_list;
                    while (parserPozriLex(f) == CSHIARKA)
                    {
                        parserLex(f);
                        ast_list->ast_list = vytvorAstList(parserAssignExp(f));
                        ast_list = ast_list->ast_list;
                    }

                    cshakajTok(RPAREN, f);
                }
            }
            else
            {
                ast = alokujAst(Prem);
                ast->un.Prem = meno;
            }

            break;
        }

        default: necshakanyyTok(tok);
    }

    return (ast);
}

static AST *parserUnaryExp(FILE *f)
{
    AST *ast;

    switch (parserPozriLex(f)) {
        case AMPERSAND:
        case MIINUS:
        case TILDE:
        case LOG_NEG: {
            int tok = parserLex(f);

            ast = alokujAst(UnOp);
            switch (tok) {
                case AMPERSAND: ast->un.UnOp.typ = UNAARNA_OP_AMPERSAND; break;
                case MIINUS: ast->un.UnOp.typ = UNAARNA_OP_NEG; break;
                case TILDE: ast->un.UnOp.typ = UNAARNA_OP_BITCOMPLEMENT; break;
                case LOG_NEG: ast->un.UnOp.typ = UNAARNA_OP_LOGNEG; break;
            }
            ast->un.UnOp.exp = parserPrimaryExp(f);

            break;
        }

        default: {
            ast = parserPrimaryExp(f);

            break;
        }
    }

    return (ast);
}

static AST *parserTerm(FILE *f)
{
    AST *ast;
    AST *faktor1;

    faktor1 = parserUnaryExp(f);
    while (parserPozriLex(f) == HVIEZDA || parserPozriLex(f) == LOMKA)
    {
        ast = alokujAst(BinOp);
        switch (parserLex(f)) {
            case HVIEZDA: ast->un.BinOp.typ = BINAARNA_OP_HVIEZDA; break;
            case LOMKA: ast->un.BinOp.typ = BINAARNA_OP_LOMKA; break;
        }
        ast->un.BinOp.exp1 = faktor1;
        ast->un.BinOp.exp2 = parserUnaryExp(f);

        faktor1 = ast;
    }

    return (faktor1);
}

static AST *parserAditiivnaExp(FILE *f)
{
    AST *ast;
    AST *term1;

    term1 = parserTerm(f);
    while (parserPozriLex(f) == PLUS || parserPozriLex(f) == MIINUS)
    {
        ast = alokujAst(BinOp);
        switch (parserLex(f)) {
            case PLUS: ast->un.BinOp.typ = BINAARNA_OP_PLUS; break;
            case MIINUS: ast->un.BinOp.typ = BINAARNA_OP_MIINUS; break;
        }
        ast->un.BinOp.exp1 = term1;
        ast->un.BinOp.exp2 = parserTerm(f);

        term1 = ast;
    }

    return (term1);
}

static AST *parserRelacshnaaExp(FILE *f)
{
    AST *ast;
    AST *exp1;
    int tok;

    exp1 = parserAditiivnaExp(f);

    tok = parserPozriLex(f);
    while (tok == LESS || tok == LESS_EQUAL
           || tok == GREATER || tok == GREATER_EQUAL)
    {
        ast = alokujAst(BinOp);
        switch (parserLex(f)) {
            case LESS: ast->un.BinOp.typ = BINAARNA_OP_LESS; break;
            case LESS_EQUAL: ast->un.BinOp.typ = BINAARNA_OP_LESS_EQUAL; break;
            case GREATER: ast->un.BinOp.typ = BINAARNA_OP_GREATER; break;
            case GREATER_EQUAL:
                ast->un.BinOp.typ = BINAARNA_OP_GREATER_EQUAL; break;
        }
        ast->un.BinOp.exp1 = exp1;
        ast->un.BinOp.exp2 = parserAditiivnaExp(f);

        exp1 = ast;

        tok = parserPozriLex(f);
    }

    return (exp1);
}

static AST *parserRovnostnaaExp(FILE *f)
{
    AST *ast;
    AST *exp1;
    int tok;

    exp1 = parserRelacshnaaExp(f);

    tok = parserPozriLex(f);
    while (tok == EQUAL || tok == NOT_EQUAL)
    {
        ast = alokujAst(BinOp);
        switch (parserLex(f)) {
            case EQUAL: ast->un.BinOp.typ = BINAARNA_OP_EQUAL; break;
            case NOT_EQUAL: ast->un.BinOp.typ = BINAARNA_OP_NOT_EQUAL; break;
        }
        ast->un.BinOp.exp1 = exp1;
        ast->un.BinOp.exp2 = parserRelacshnaaExp(f);

        exp1 = ast;

        tok = parserPozriLex(f);
    }

    return (exp1);
}

static AST *parserLogickeeAExp(FILE *f)
{
    AST *ast;
    AST *exp1;
    int tok;

    exp1 = parserRovnostnaaExp(f);

    tok = parserPozriLex(f);
    while (tok == LOG_AND)
    {
        ast = alokujAst(BinOp);
        parserLex(f);
        ast->un.BinOp.typ = BINAARNA_OP_LOG_AND;
        ast->un.BinOp.exp1 = exp1;
        ast->un.BinOp.exp2 = parserRovnostnaaExp(f);

        exp1 = ast;

        tok = parserPozriLex(f);
    }

    return (exp1);
}

static AST *parserLogickeeAleboExp(FILE *f)
{
    AST *ast;
    AST *exp1;
    int tok;

    exp1 = parserLogickeeAExp(f);

    tok = parserPozriLex(f);
    while (tok == LOG_OR)
    {
        ast = alokujAst(BinOp);
        parserLex(f);
        ast->un.BinOp.typ = BINAARNA_OP_LOG_OR;
        ast->un.BinOp.exp1 = exp1;
        ast->un.BinOp.exp2 = parserLogickeeAExp(f);

        exp1 = ast;

        tok = parserPozriLex(f);
    }

    return (exp1);
}

static AST *parserKondicionaalnaExp(FILE *f)
{
    AST *ast;
    AST *exp1;

    exp1 = parserLogickeeAleboExp(f);

    if (parserPozriLex(f) == OTAAZNIK)
    {
        ast = alokujAst(TernOp);
        parserLex(f);
        ast->un.TernOp.exp1 = exp1;
        ast->un.TernOp.exp2 = parserExp(f);
        cshakajTok(KOLON, f);
        ast->un.TernOp.exp3 = parserKondicionaalnaExp(f);

        return (ast);
    }

    return (exp1);
}

static AST *parserAssignExp(FILE *f)
{
    AST *ast;
    AST *exp1;

    exp1 = parserKondicionaalnaExp(f);

    if (parserPozriLex(f) == ASSIGN)
    {
        /* +++Pokrocshilaa logika, aby len unaarna a pod. exp. fungovala. */
        ast = alokujAst(Assign);
        parserLex(f);
        ast->un.Assign.exp1 = exp1;
        ast->un.Assign.exp2 = parserAssignExp(f);
    }
    else
    {
        return (exp1);
    }

    return (ast);
}

static AST *parserExp(FILE *f)
{
    return (parserAssignExp(f));
}

static AST *parserStatement(FILE *f)
{
    AST *ast;
    int tok;

    tok = parserPozriLex(f);
    switch (tok) {
        case IF: {
            parserLex(f);

            ast = alokujAst(If);

            cshakajTok(LPAREN, f);
            ast->un.If.exp = parserExp(f);
            cshakajTok(RPAREN, f);

            ast->un.If.statement = parserStatement(f);

            if (parserPozriLex(f) == ELSE)
            {
                parserLex(f);
                ast->un.If.else_statement = parserStatement(f);
            }
            else ast->un.If.else_statement = NULL;

            break;
        }

        case RETURN: {
            parserLex(f);

            ast = alokujAst(Return);
            ast->un.Return.exp = parserExp(f);

            cshakajTok(SEMIKOLON, f);
            break;
        }

        case LBRACE: {
            parserLex(f);

            ast = alokujAst(Blok);
            if (parserPozriLex(f) == RBRACE)
            {
                ast->un.Blok.blok_item_list = NULL;
            }
            else
            {
                AST_LIST *ast_list;

                ast_list = vytvorAstList(parserBlokItem(f));
                ast->un.Blok.blok_item_list = ast_list;
                while (parserPozriLex(f) != RBRACE)
                {
                    ast_list->ast_list = vytvorAstList(parserBlokItem(f));
                    ast_list = ast_list->ast_list;
                }
            }

            cshakajTok(RBRACE, f);
            break;
        }

        case FOR: {
            parserLex(f);

            ast = alokujAst(For);

            cshakajTok(LPAREN, f);
            switch (parserPozriLex(f)) {
                case INT: {
                    ast->un.For.dob_exp1 = parserDekl(f);

                    break;
                }

                case SEMIKOLON: {
                    ast->un.For.dob_exp1 = NULL;
                    parserLex(f);

                    break;
                }

                default: {
                    ast->un.For.dob_exp1 = parserExp(f);
                    cshakajTok(SEMIKOLON, f);

                    break;
                }
            }

            if (parserPozriLex(f) == SEMIKOLON)
            {
                /* Vynechanaa riadiaca exp.
                 * je nahradenaa nenulovou konshtantnou exp.. */
                ast->un.For.exp2 = alokujAst(Konshtanta);
                ast->un.For.exp2->un.Konshtanta = 1;
                parserLex(f);
            }
            else
            {
                ast->un.For.exp2 = parserExp(f);
                cshakajTok(SEMIKOLON, f);
            }

            if (parserPozriLex(f) == RPAREN)
            {
                ast->un.For.dob_exp3 = NULL;
                parserLex(f);
            }
            else
            {
                ast->un.For.dob_exp3 = parserExp(f);
                cshakajTok(RPAREN, f);
            }

            ast->un.For.statement = parserStatement(f);

            break;
        }

        case WHILE: {
            parserLex(f);

            ast = alokujAst(While);

            cshakajTok(LPAREN, f);
            ast->un.While.exp = parserExp(f);
            cshakajTok(RPAREN, f);

            ast->un.While.statement = parserStatement(f);

            break;
        }

        case DO: {
            parserLex(f);

            ast = alokujAst(Do_While);

            ast->un.Do_While.statement = parserStatement(f);

            cshakajTok(WHILE, f);
            cshakajTok(LPAREN, f);
            ast->un.Do_While.exp = parserExp(f);
            cshakajTok(RPAREN, f);

            cshakajTok(SEMIKOLON, f);
            break;
        }

        case CONTINUE: {
            parserLex(f);

            ast = alokujAst(Continue);

            cshakajTok(SEMIKOLON, f);
            break;
        }

        case BREAK: {
            parserLex(f);

            ast = alokujAst(Break);

            cshakajTok(SEMIKOLON, f);
            break;
        }

        case SEMIKOLON: {
            parserLex(f);

            ast = NULL;

            break;
        }

        default: {
            ast = parserExp(f);

            cshakajTok(SEMIKOLON, f);
            break;
        }
    }

    return (ast);
}

static AST *parserDekl(FILE *f)
{
    AST *ast;

    cshakajTok(INT, f);
    ast = alokujAst(Dekl);

    cshakajTok(IDENT, f);
    ast->un.Dekl.meno = identMeno();

    if (parserPozriLex(f) == ASSIGN)
    {
        parserLex(f);
        ast->un.Dekl.dob_exp = parserExp(f);
    }
    else
    {
        ast->un.Dekl.dob_exp = NULL;
    }

    cshakajTok(SEMIKOLON, f);

    return (ast);
}

static AST *parserBlokItem(FILE *f)
{
    switch (parserPozriLex(f)) {
        case INT: {
            return (parserDekl(f));
        }

        default: return(parserStatement(f));
    }
}

static AST *parserParameter(FILE *f)
{
    AST *ast;

    cshakajTok(INT, f);
    ast = alokujAst(Parameter);

    cshakajTok(IDENT, f);
    ast->un.Parameter.meno = identMeno();

    return (ast);
}

AST *parserExternalDeclaration(FILE *f)
{
    AST *ast;
    char *meno;

    if (parserLex(f) != INT) parserZlyhanie("nie INT\n");
    if (parserLex(f) != IDENT) parserZlyhanie("nie IDENT\n");
    meno = identMeno();

    if (parserPozriLex(f) == LPAREN)
    {
        parserLex(f);

        ast = alokujAst(Funkcia);

        ast->un.Funkcia.meno = meno;
        if (parserPozriLex(f) == RPAREN)
        {
            ast->un.Funkcia.parameter_list = NULL;
        }
        else
        {
            AST_LIST *ast_list;

            ast_list = vytvorAstList(parserParameter(f));
            ast->un.Funkcia.parameter_list = ast_list;
            while (parserPozriLex(f) == CSHIARKA)
            {
                parserLex(f);
                ast_list->ast_list = vytvorAstList(parserParameter(f));
                ast_list = ast_list->ast_list;
            }
        }
        cshakajTok(RPAREN, f);

        if (parserLex(f) != LBRACE) parserZlyhanie("nie LBRACE\n");
        if (parserPozriLex(f) == RBRACE)
        {
            ast->un.Funkcia.blok_item_list = NULL;
        }
        else
        {
            AST_LIST *ast_list;

            ast_list = vytvorAstList(parserBlokItem(f));
            ast->un.Funkcia.blok_item_list = ast_list;
            while (parserPozriLex(f) != RBRACE)
            {
                ast_list->ast_list = vytvorAstList(parserBlokItem(f));
                ast_list = ast_list->ast_list;
            }
        }
        parserLex(f);
    }
    else
    {
        ast = alokujAst(ExternalDekl);

        ast->un.ExternalDekl.meno = meno;

        cshakajTok(SEMIKOLON, f);
    }

    return (ast);
}

AST *parserTranslationUnit(FILE *f)
{
    AST *ast;
    AST *translationUnit = NULL;

    if (parserPozriLex(f) == END_OF_FILE)
    {
        parserZlyhanie("Praazdna TranslationUnit\n");
    }

    while (parserPozriLex(f) != END_OF_FILE)
    {
        ast = alokujAst(TranslationUnit);

        ast->un.TranslationUnit.translationUnit = translationUnit;
        ast->un.TranslationUnit.funkcia = parserExternalDeclaration(f);

        translationUnit = ast;
    }

    return (ast);
}
