/*
 * RSH - Robte si hocičo
 *
 * Ktokoľvek a čokoľvek smie kopírovať, upravovať, zverejňovať,
 * používať, kompilovať, predávať a šíriť toto dielo
 * a všetky jeho časti v akejkoľvek podobe na akékoľvek účely,
 * komerčné a nekomerčné, bez akýchkoľvek obmedzení,
 * bez dodržiavania akýchkoľvek podmienok
 * a akýmikoľvek prostriedkami.
 *
 * Jeder und alles darf kopieren, bearbeiten, veröffentlichen,
 * verwenden, kompilieren, verkaufen und verbreiten dieses Werk
 * und alle seine Teile in irgendeiner Form für irgendeinen Zweck,
 * kommerziell und nichtkommerziell, ohne Beschränkungen,
 * ohne irgendwelche Bedingungen einzuhalten
 * und mit allen Mitteln.
 *
 * Anyone and anything may copy, edit, publish,
 * use, compile, sell and distribute this work
 * and all its parts in any form for any purpose,
 * commercial and non-commercial, without any restrictions,
 * without complying with any conditions
 * and by any means.
 */

#ifndef PARSER_H
#define PARSER_H

#include <stdio.h>

#include "konshtanty.h"

enum {UNAARNA_OP_NEG, UNAARNA_OP_BITCOMPLEMENT, UNAARNA_OP_LOGNEG,
      UNAARNA_OP_AMPERSAND};

enum {BINAARNA_OP_HVIEZDA, BINAARNA_OP_LOMKA,
      BINAARNA_OP_PLUS, BINAARNA_OP_MIINUS,
      BINAARNA_OP_LESS, BINAARNA_OP_LESS_EQUAL,
      BINAARNA_OP_GREATER, BINAARNA_OP_GREATER_EQUAL,
      BINAARNA_OP_EQUAL, BINAARNA_OP_NOT_EQUAL,
      BINAARNA_OP_LOG_OR,
      BINAARNA_OP_LOG_AND};

typedef struct _ast {
    enum {TranslationUnit,
          Funkcia, Parameter, ExternalDekl,
          If, Return, Blok, For, While, Do_While, Continue, Break,
          Dekl,
          UnOp, BinOp, TernOp, Assign, Volanie, Prem, Konshtanta} typ;
    union {
        struct {
            struct _ast *translationUnit;
            struct _ast *funkcia;
        } TranslationUnit;

        struct {
            char *meno;
            struct _ast_list *parameter_list;
            struct _ast_list *blok_item_list;
        } Funkcia;

        struct {
            char *meno;
        } Parameter;

        struct {
            char *meno;
        } ExternalDekl;

        struct {
            struct _ast *exp;
            struct _ast *statement;
            struct _ast *else_statement;
        } If;

        struct {
            struct _ast *exp;
        } Return;

        struct {
            struct _ast_list *blok_item_list;
        } Blok;

        struct {
            struct _ast *dob_exp1;
            struct _ast *exp2;
            struct _ast *dob_exp3;
            struct _ast *statement;
        } For;

        struct {
            struct _ast *exp;
            struct _ast *statement;
        } While;

        struct {
            struct _ast *statement;
            struct _ast *exp;
        } Do_While;

        struct {
            char *meno;
            struct _ast *dob_exp;
        } Dekl;

        struct {
            int typ;
            struct _ast *exp;
        } UnOp;

        struct {
            int typ;
            struct _ast *exp1;
            struct _ast *exp2;
        } BinOp;

        struct {
            struct _ast *exp1;
            struct _ast *exp2;
            struct _ast *exp3;
        } TernOp;

        struct {
            struct _ast *exp1;
            struct _ast *exp2;
        } Assign;

        struct {
            char *meno;
            struct _ast_list *assign_exp_list;
        } Volanie;

        char *Prem;

        int Konshtanta;
    } un;
} AST;

typedef struct _ast_list {
    struct _ast *ast;
    struct _ast_list *ast_list;
} AST_LIST;

AST *parserTranslationUnit(FILE *f);

#endif /* PARSER_H */
